Import Irrlicht.Core
'Import "D:\minib3d\backup\MiniB3D-v030\minib3d.bmx"
Import SiDesign.MiniB3D

' -------------------
' new plan.  this has to be a new animated mesh type.  examine the animation functions and see
' what is needed.  intital plan: based on current timeframe set the verts of each surf the surf_list 
' to the current frame.
' NOTE: need a frames per second setting.
' NOTE: need a frame loop setting
' NOTE: potentially need all functionality from IAnimatedSceneNode
' TODO: create optional cached version.  preloads and stores each frame in a surf_list.
' TODO: change preload to cachemode.  have preload all, load save, and load always
' TODO: instead of loading textures from zip, save to temp and then load as normal
'
Rem
AnimateMD2 md2[,mode][,speed#][,first_frame][,last_frame][,transition#]
Parameters:
md2 - md2 handle

mode (optional) - mode of animation
0: stop animation
1: loop animation (default)
2: ping-pong animation
3: one-shot animation

speed# (optional) - speed of animation. Defaults to 1.
first_frame (optional) - first frame of animation. Defaults to 1.
last_frame# (optional) - last frame of animation. Defaults to last frame of all md2 animations.
transition# (optional) - smoothness of transition between last frame shown of previous animation and first frame of next animation. Defaults to 0.
EndRem
' thought: can we use the verts/tris directly from Irrlicht and keep the loader open?
' 	- have a C routine that returns the verts and tris and use this to populate a surface on the fly
'	- still need to setup the surface brushes
'	- averaging 50-60 msecs per mesh buffer
' each frame needs its own verts, brushes
' can update the verts by overriding the update and setting them before the super
'
' each MeshBuffer needs its own surface.
' each frame, for each surface needs the verts and brushes.
' -------------------
' -------------------

Type IrrlichtMeshLoader
	Global mesh_list:TMap
	Global entity_list:TMap	
	Global device:IrrlichtDevice
	Global smgr:ISceneManager
		
	Function LoadMesh:IrrlichtMeshFrameList(file:String,bPreLoad:Int=False)
		If Not mesh_list Then mesh_list=CreateMap()		
		If Not entity_list Then entity_list=CreateMap()
		Local irrMesh:IrrlichtMeshFrameList=Null
		
		If InitLoader()		
			smgr=device.getSceneManager()
				
			If smgr And smgr.IsValid() Then
				'DebugStop
				irrMesh=IrrlichtMeshFrameList.Create(smgr.getMesh(file),bPreLoad)		
				mesh_list.Insert(file,irrMesh)
				entity_list.Insert(irrMesh,CreateList())
			EndIf
		EndIf
		
		Return irrMesh
	EndFunction

	Function AddZipFileArchive(filename:String)
		If InitLoader()
			device.getFileSystem().addZipFileArchive(filename)
		EndIf
	EndFunction

	Function InitLoader:Int()
		If Not device Or Not device.IsValid() Then device=IrrlichtDevice.Create(EDT_NULL,_DIMENSION2DI(0,0))
		Return device And device.IsValid()
	EndFunction

	Function RemoveMesh(file:String)
		If mesh_list Then mesh_list.Remove(file)
		If mesh_list.IsEmpty() Then 
			mesh_list=Null
			If device And device.IsValid() Then
				device.CloseDevice()
				device.drop()	
			EndIf
		EndIf
	EndFunction
	
	' updates 
	Function UpdateWorld()
		' for each Irrlicht mesh animate
		For Local mesh:IrrlichtMesh=EachIn IrrlichtMesh.entity_list
			mesh.AnimateMesh()
		Next
	EndFunction	
EndType

' things to improve:
' - speed up the loading by better using the available functions
' - speed up loading by directly adding indicies
' - all else fails, cache the frames
' 

Type IrrlichtMeshTriangleInfo
	' replication vars.  used once per surface.
	Field no_tris=0
	Field tris:Short[0]
	Field tri_array_size=1
	Field brush:TBrush
	
	Method CopyToSurface(surf:TSurface)
		surf.no_tris=no_tris
		surf.tris=tris[..]
		surf.tri_array_size=tri_array_size
		If brush<>Null
			surf.brush=brush
		EndIf
	EndMethod
	
	Method CopyFromSurface(surf:TSurface)
		no_tris=surf.no_tris
		tris=surf.tris[..]
		tri_array_size=surf.tri_array_size
		If surf.brush<>Null
			brush=surf.brush.Copy()
		EndIf
	EndMethod	
EndType

Type IrrlichtMeshFrameList
	Field mesh:IAnimatedMesh
	Field frameCnt:Int
	Field anim_cache:Int=False
	Field anim_frames:TList[]
	Field anim_tri_cache:IrrlichtMeshTriangleInfo[]
	
	Method CreateMesh:IrrlichtMesh()
		Return IrrlichtMesh.Create(Self)
	EndMethod
	
	Function Create:IrrlichtMeshFrameList(iMesh:IAnimatedMesh,bPreLoad:Int=False)
		Local frames:IrrlichtMeshFrameList=Null
		' build the mesh and process the first frame
		If iMesh And iMesh.IsValid() Then
			frames=New IrrlichtMeshFrameList
			
			frames.mesh=iMesh
			frames.frameCnt=iMesh.getFrameCount()
			If frames.frameCnt=1 Then bPreload=True
			frames.anim_cache=bPreLoad
						
			If bPreLoad Then
				frames.anim_frames=New TList[frames.frameCnt]
			Else
				frames.anim_frames=New TList[1]
			EndIf
						
			For Local frame:Int=0 Until frames.frameCnt
				If Not frames.anim_frames[frame] Then frames.anim_frames[frame]=CreateList()
				frames.BuildSurfaces(frames.anim_frames[frame],frame)
				If Not bPreLoad Then Exit
			Next
			
			If bPreLoad Then IrrlichtMeshLoader.smgr.getMeshCache().removeMesh(iMesh)
		EndIf
		Return frames
	EndFunction
	
	Method ApplyFrame(mesh:IrrlichtMesh,frame:Int=0)
		' check if loaded.  if not load it.
		If anim_cache Then 
			mesh.surf_list=anim_frames[frame]
			' set the tris
			Local cnt:Int=0
			For Local surf:TSurface=EachIn mesh.surf_list
				anim_tri_cache[cnt].CopyToSurface(surf)
				cnt:+1
			Next
		Else				
			BuildSurfaces(anim_frames[0],frame)
			mesh.surf_list=anim_frames[0]
		EndIf
		mesh.no_surfs=mesh.surf_list.Count()
	EndMethod
	
	Method BuildSurfaces(surfaces:TList,frame:Int=0)

		'DebugLog("Building frame: "+frame)
		'Local start:Int=MilliSecs()
		'DebugLog("surfaces start: "+GCMemAlloced())

		' clear out any existing surfaces
		surfaces.Clear()
		
		' get the # of surfaces for this frame
		Local surfCnt:Int=mesh.getMesh(frame).getMeshBufferCount()
		
		If Not anim_tri_cache Then anim_tri_cache=New IrrlichtMeshTriangleInfo[surfCnt]
		
		' each mesh buffer for the frame is a surface
		For Local surfIdx:Int=0 Until surfCnt			
			
			' get the current buffer
			Local mbuf:IMeshBuffer=mesh.getMesh(frame).getMeshBuffer(surfIdx)
			
			' add verts to a new surface on the mesh (one surface per frame)
			Local new_surf:TSurface=CreateSurfaceFromMeshBuffer(mbuf,(anim_tri_cache[surfIdx]<>Null))
			surfaces.AddLast(new_surf)												
			
			' setup joints									

			' setup the triangle/brush cache
			If Not anim_tri_cache[surfIdx] Then
				' setup textures/brushes
				Local brush:TBrush=CreateBrushFromMaterial(mbuf.getMaterial(),mbuf.getVertexType())

				anim_tri_cache[surfIdx]=New IrrlichtMeshTriangleInfo
				anim_tri_cache[surfIdx].CopyFromSurface(new_surf)
				anim_tri_cache[surfIdx].brush=brush
			Else
				' copy the tris to the surface
				anim_tri_cache[surfIdx].CopyToSurface(new_surf)
			EndIf
		Next
		'DebugLog("surfaces end: "+(MilliSecs()-start)+" | "+GCMemAlloced())
	EndMethod
	
	Method CreateSurfaceFromMeshBuffer:TSurface(mbuf:IMeshBuffer,bNoTris:Int=False)
		Local vertType:Int=mbuf.getVertexType()
		
		'Local start:Int=MilliSecs()
		'DebugLog("start: "+start)

		Local v_surf:TSurface=New TSurface
		Local vertCnt:Int=mbuf.getVertexCount()

		' setup the surface
		v_surf.vert_coords=v_surf.vert_coords[..vertCnt*3]
		v_surf.vert_tex_coords0=v_surf.vert_tex_coords0[..vertCnt*2]
		v_surf.vert_norm=v_surf.vert_norm[..vertCnt*3]
		v_surf.vert_col=v_surf.vert_col[..vertCnt*4]
		v_surf.no_verts=vertCnt
		v_surf.vert_array_size=vertCnt

		' prepopulate 
		If vertType=EVT_STANDARD
			mbuf.fillS3DVertexVerticesPtrs(v_surf.vert_coords,v_surf.vert_norm,v_surf.vert_tex_coords0,v_surf.vert_col,True)
		ElseIf vertType=EVT_2TCOORDS
			'DebugLog("2T")
			v_surf.vert_tex_coords1=v_surf.vert_tex_coords1[..vertCnt*2]
			mbuf.fillS3DVertex2TCoordsVerticesPtrs(v_surf.vert_coords,v_surf.vert_norm,v_surf.vert_tex_coords0,v_surf.vert_col,v_surf.vert_tex_coords1,True)
		ElseIf vertType=EVT_TANGENTS
			Local tangents:Float[vertCnt*3]
			Local binormals:Float[vertCnt*3]
			mbuf.fillS3DVertexTangentsVerticesPtrs(v_surf.vert_coords,v_surf.vert_norm,v_surf.vert_tex_coords0,v_surf.vert_col,tangents,binormals,True)
		
			' TODO: not sure what to do with the binormals and tangents		
		EndIf
		
		If Not bNoTris Then			
			v_surf.no_tris=mbuf.getIndexCount()
			v_surf.tri_array_size=v_surf.no_tris
			v_surf.tris=v_surf.tris[..v_surf.no_tris*3]				
			IrrScene_IMeshBuffer_fillIndicesPtr(mbuf.handle,v_surf.tris,True)
		EndIf
			
		'DebugLog("end: "+(MilliSecs()-start))
		Return v_surf
	EndMethod

	Method CreateBrushFromMaterial:TBrush(material:SMaterial,vertType:Int)
		Local brush:TBrush=New TBrush
		Local mtype:Int=material.getMaterialType()
		Local bUseTCoords2:Int=(vertType=EVT_2TCOORDS And (mtype=EMT_LIGHTMAP Or mtype=EMT_LIGHTMAP_ADD Or mtype=EMT_LIGHTMAP_M2 Or mtype=EMT_LIGHTMAP_M4))
		Local bBlendAdd:Int=(mtype=EMT_LIGHTMAP_ADD)
		Local bBlendMult:Int=(mtype=EMT_LIGHTMAP_M2 Or mtype=EMT_LIGHTMAP_M4 Or mtype=EMT_LIGHTMAP_LIGHTING_M2 Or mtype=EMT_LIGHTMAP_LIGHTING_M4)
		Local flags:Int=0
								
		' TODO: set FX
		'If bBlendAdd Then brush.BrushBlend(3)
		'If bBlendMult Then brush.BrushBlend(2)
		
		If Not material.getLighting() Then flags:+1 ' fullbright
		If Not material.getGouraudShading() Then flags:+4 ' flat shaded
		If Not material.getFogEnable() Then flags:+8 ' fog disable
		If Not material.getBackfaceCulling() Then flags:+16 ' backface culling disable
		
		brush.BrushFX(flags)
		
		Rem
		TextureFilter
		1: Color
2: Alpha
4: Masked
8: Mipmapped
16: Clamp U
32: Clamp V
64: Spherical environment map
128: Cubic environment mapping
256: Store texture in vram
512: Force the use of high color textures
		EndRem
		' load textures
		For Local cnt:Int=0 Until MATERIAL_MAX_TEXTURES
			Local tex_temp:ITexture = material.getTexture(cnt)
			If tex_temp Then 
				Local filename:String=tex_temp.getName()
				Local bCleanUp:Int=False
				If filename<>"" Then
					If FileType(filename)=0 Then
						filename=CreateFileFromTexture(material.getTexture(cnt))
						bCleanUp=True
					EndIf
					
					If FileType(filename)<>0 Then
						Local tex:TTexture=TTexture.LoadTexture(filename)
						' see if this is the lightmap
						If cnt=1 And bUseTCoords2 Then tex.TextureCoords(1)
						brush.BrushTexture(tex,cnt)
						DebugLog("applying texture: "+filename)
					EndIf
					
					If bCleanUp Then
						DeleteFile(filename)
					EndIf
				EndIf
			EndIf
			material.setTexture(cnt,Null)
		Next
				
		Return brush
	EndMethod

	Method CreateFileFromTexture:String(texture:ITexture)
		Local filename:String=Null
		Local data:Byte Ptr=texture.lock()

		If data Then
			filename=texture.getName().Replace(".jpg",".png")
			filename = filename.Replace(ExtractDir(filename) + "/", "")
			Local image:IImage=IrrlichtMeshLoader.device.getVideoDriver().createImageFromData(texture.getColorFormat(),texture.getSize(),data,False)
			texture.unlock()
			data=Null
			IrrlichtMeshLoader.device.getVideoDriver().writeImageToFile(image,filename)
			image.drop()			
		EndIf
		
		Return filename
	EndMethod
EndType

Type IrrlichtMesh Extends TMesh

	Global entity_list:TList=CreateList()

	Field currentFrame:Int
	Field currentFrameTime:Int
	Field lastFrame:Int
	Field frames:IrrlichtMeshFrameList=Null
	Field framesPerMilliSec:Float=.03

	Method SetFramesPerSec(framesPerSec:Int=30)
		framesPerMilliSec=Float(framesPerSec)/Float(1000)
	EndMethod

	Method GetFramesPerSec:Int()
		Return (framesPerMilliSec*1000)
	EndMethod

	' sets a new frame
	Method SetFrame(frame:Int)
		If frame<0 Then frame=(frames.frameCnt-1)
		If frame>=frames.frameCnt Then frame=0 
		currentFrame=frame

		' see if we need to changeup the frames
		If currentFrame<>lastFrame Or Not surf_list Then 
			frames.ApplyFrame(Self,currentFrame)
			lastFrame=currentFrame
		EndIf
		
		currentFrameTime=MilliSecs()						
	EndMethod
		
	Method GetFrame:Int()
		Return currentFrame
	EndMethod
	
	Method SetFrameList(mesh_frames:IrrlichtMeshFrameList)
		' first, unlink any previous
		If frames Then 
			TList(IrrlichtMeshLoader.entity_list.ValueForKey(frames)).Remove(Self)		
		EndIf		
		frames=mesh_frames		
		' bind the mesh to the mesh frames
		TList(IrrlichtMeshLoader.entity_list.ValueForKey(mesh_frames)).AddLast(Self)		
		' setup the initial frame
		mesh_frames.ApplyFrame(Self,0)		
	EndMethod
	
	Function Create:IrrlichtMesh(mesh_frames:IrrlichtMeshFrameList)
		Local irrMesh:IrrlichtMesh=New IrrlichtMesh
		irrMesh.SetFrameList(mesh_frames)
		' setup for rendering
		irrMesh.EntityListAdd(TEntity.entity_list)
		irrMesh.EntityListAdd(IrrlichtMesh.entity_list)
		Return irrMesh
	EndFunction
	
	Method New()
		currentFrame=0
		lastFrame=0
		currentFrameTime=MilliSecs()
	EndMethod
	
	Method AnimateMesh()
		' see if we need to increment the frame
		Local skip_frames:Int=(MilliSecs()-currentFrameTime)*framesPerMilliSec
		'DebugLog((MilliSecs()-currentFrameTime)*framesPerMilliSec)
		If skip_frames>0			
			SetFrame(GetFrame()+skip_frames)
		EndIf
	EndMethod
	
	Rem
	Method Update()
		' build this frame
		Super.Update()
	EndMethod				
	EndRem
EndType


